/*
 * basis.c - implementation of the user defined type
 */

#include "postgres.h"
#include "funcapi.h"
#include "access/heapam.h"
#include "basis.h"
#include "model.h"

float8 ltqnorm(float8 p);

/* Reusable buffers for handling PDF_INIT instances. */
static char BUF1[NORM_BUF_SIZE];
static char BUF2[NORM_BUF_SIZE];

/* If PDF_INIT, construct a temp PDF_CERT or PDF_NORM. */
#define check_init(pdf, buf) \
	if ((pdf)->type == PDF_INIT) \
		(pdf) = get_init_pdf(PDF_BASIS((pdf)), (buf))

/*----------------------------------------------------------------------------*\
  PDF Support Functions
\*----------------------------------------------------------------------------*/

/*
 * Gets the expected value of the distribution.
 */
float4 c_expect(PDF *pdf, bool *isnull) {
	check_init(pdf, BUF1);
	switch (pdf->type) {
	case PDF_ZERO:
		*isnull = true;
		return 0;
	case PDF_CERT:
	case PDF_NORM:
		*isnull = false;
		return pdf->data[0];
	default:
		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
						errmsg("c_expect() - unsupported pdf type")));
		return 0; /* keep compiler quiet */
	}
}

/*
 * Gets the standard deviation of the distribution.
 */
float4 c_stddev(PDF *pdf, bool *isnull) {
	check_init(pdf, BUF1);
	switch (pdf->type) {
	case PDF_ZERO:
		*isnull = true;
		return 0;
	case PDF_CERT:
		*isnull = false;
		return 0;
	case PDF_NORM:
		*isnull = false;
		return pdf->data[1];
	default:
		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
						errmsg("c_stddev() - unsupported pdf type")));
		return 0; /* keep compiler quiet */
	}
}

/*
 * Gets an interval that contains the given minimum probability mass.
 * Returns the actual mass contained in the resulting interval.
 */
float4 c_confint(PDF *pdf, float4 mass, float4 *out_minv, float4 *out_maxv) {
	check_init(pdf, BUF1);
	switch (pdf->type) {
	case PDF_ZERO:
		*out_minv = 0;
		*out_maxv = 0;
		return 0;
	case PDF_CERT:
		*out_minv = pdf->data[0];
		*out_maxv = pdf->data[0];
		return 1;
	case PDF_NORM: {
		/* get and scale the inverse CDF */
		float4 bias = pdf->data[1] * ltqnorm(0.5 * (1 + mass));
		*out_minv = pdf->data[0] - bias;
		*out_maxv = pdf->data[0] + bias;
		return mass;
	}
	default:
		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
						errmsg("c_confint() - unsupported pdf type")));
		return 0; /* keep compiler quiet */
	}
}

/*
 * Calculates the JS divergence between two normal distributions.
 */
float4 c_diverge(PDF *p1, PDF *p2) {
	float4 m1, s1, m2, s2;
	check_init(p1, BUF1);
	check_init(p2, BUF2);
	if (p1->type == PDF_ZERO || p2->type == PDF_ZERO) {
		return INFINITY;
	}
	if (p1->type != PDF_NORM || p2->type != PDF_NORM) {
		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
						errmsg("c_diverge() - unsupported pdf type")));
	}
	m1 = p1->data[0];
	s1 = p1->data[1];
	m2 = p2->data[0];
	s2 = p2->data[1];
	return log(s2 / s1) + (s1*s1 + (m1 - m2)*(m1 - m2)) / (2 * s2*s2) - 0.5;
}

/*----------------------------------------------------------------------------*\
  C Implementation
\*----------------------------------------------------------------------------*/

/*
 * Input function.
 */
static Basis *c_basis_in(char *str) {
	/* TODO: syntax for additional state (e.g., predictors) */
	static char *error = "direct input not yet implemented; use new_basis()";
	static char *hint = "finish basis_in and basis_out before using pg_dump";
	ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
					errmsg("%s", error), errhint("%s", hint)));
	return NULL; /* keep compiler quiet */
}

/*
 * Output function.
 */
static char *c_basis_out(Basis *basis) {
	int32 i;
	PDF *pdf = &basis->pdf;
	int32 maxlen = (13 + 13 + 2 + 2 + 7) + (2 + 7 + 7 + 13 * pdf->length);
	char *str = (char *) palloc(maxlen);
	char *pos = str;

	pos += sprintf(pos, "%g;%g|%d;%d;%d|%c;%d;%d|", basis->initval,
			basis->istddev, basis->hasinit, basis->suspect, basis->round,
			pdf->type, pdf->model, pdf->length);
	for (i = 0; i < pdf->length; i++) {
		pos += sprintf(pos, "%g,", pdf->data[i]);
	}
	pos--;
	*pos = '\0'; /* overwrite extra delimiter */

	return str;
}

/*
 * Returns a relational view of the structure's internals.
 */
static HeapTupleHeader c_get_info(Basis *basis, TupleDesc tupdesc) {
	float4 minv, maxv;
	HeapTuple tuple;
	Datum values[13];
	bool isnull[13] = {0};

	/* compute the pdf statistics */
	PDF *pdf = &basis->pdf;
	float4 ev = c_expect(pdf, &isnull[9]);
	float4 sd = c_stddev(pdf, &isnull[10]);
	float4 ci = c_confint(pdf, ERROR_THOLD, &minv, &maxv);
	text *pdftype = palloc0(VARHDRSZ + 1);
	SET_VARSIZE(pdftype, VARHDRSZ + 1);
	pdftype->vl_dat[0] = pdf->type;

	/* construct and return the resulting tuple */
	values[0] = Int32GetDatum(VARSIZE(basis));
	values[1] = Float4GetDatum(basis->initval);
	values[2] = Float4GetDatum(basis->istddev);
	values[3] = BoolGetDatum(basis->hasinit);
	values[4] = BoolGetDatum(basis->suspect);
	values[5] = Int16GetDatum(basis->round);
	values[6] = PointerGetDatum(pdftype);
	values[7] = Int16GetDatum(pdf->model);
	values[8] = Int16GetDatum(pdf->length);
	values[9] = Float4GetDatum(ev);
	values[10] = Float4GetDatum(sd);
	if (ci > 0) {
		values[11] = Float4GetDatum(minv);
		values[12] = Float4GetDatum(maxv);
	} else {
		isnull[11] = true;
		isnull[12] = true;
	}
	tuple = heap_form_tuple(tupdesc, values, isnull);
	return tuple->t_data;
}

/*
 * Constructor that sets the initial values (given from SQL).
 */
static Basis *c_new_basis(float4 initval, float4 istddev, bool hasinit) {
	Basis *basis = palloc0(sizeof(Basis));
	SET_VARSIZE(basis, sizeof(Basis));
	if (hasinit) {
		basis->initval = initval;
		basis->hasinit = hasinit;
		basis->pdf.type = PDF_INIT;
	} else {
		basis->pdf.type = PDF_ZERO;
	}
	basis->istddev = istddev;
	basis->pdf.model = -1;
	return basis;
}

/*
 * Copy constructor that appends the basis with the updated pdf.
 */
Basis *update_basis(Basis *basis, PDF *pdf) {
	int32 elen, size;
	Basis *ret;

	/* calculate the length of extra data */
	elen = pdf->length * 4;
	if (g_version & 4 && pdf->model > -1) {
		elen += g_nvar * 12; /* see erace_var */
	}

	/* allocate and copy the basis */
	size = sizeof(Basis) + elen;
	ret = palloc(size);
	MemSet(ret, 0, sizeof(Basis) - sizeof(PDF));
	SET_VARSIZE(ret, size);
	ret->initval = basis->initval;
	ret->istddev = basis->istddev;
	ret->hasinit = basis->hasinit;
	ret->suspect = basis->suspect;
	ret->round = g_round;
	g_updcnt++;

	/* copy the variable-length PDF */
	size = sizeof(PDF) + elen;
	memcpy(&ret->pdf, pdf, size);
	return ret;
}

/*
 * Copy constructor that reverts the basis to its initial pdf.
 */
Basis *revert_basis(Basis *basis) {
	PDF pdf = {PDF_INIT, -1, 0};
	return update_basis(basis, &pdf);
}

/*
 * Constructs a temporary pdf over the initial basis values, using the given
 * buffer to avoid unnecessary palloc/pfree calls.
 */
PDF *get_init_pdf(Basis *basis, char *buf) {
	PDF *pdf = (PDF *) buf;
	MemSet(pdf, 0, NORM_BUF_SIZE);
	pdf->model = -1;
	if (basis->hasinit) {
		pdf->data[0] = basis->initval;
		if (basis->istddev == 0) {
			pdf->type = PDF_CERT;
			pdf->length = 1;
		} else {
			pdf->type = PDF_NORM;
			pdf->length = 2;
			pdf->data[1] = basis->istddev;
		}
	} else {
		pdf->type = PDF_ZERO;
		pdf->length = 0;
	}
	return pdf;
}

/*----------------------------------------------------------------------------*\
  Postgres V1 Function Prototypes
\*----------------------------------------------------------------------------*/

Datum basis_in(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(basis_in);

Datum basis_out(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(basis_out);

Datum new_basis(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(new_basis);

Datum get_info(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(get_info);

Datum expect(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(expect);

Datum stddev(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(stddev);

Datum confint(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(confint);

Datum diverge(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(diverge);

/*----------------------------------------------------------------------------*\
  SQL Interface (Wrappers)
\*----------------------------------------------------------------------------*/

Datum basis_in(PG_FUNCTION_ARGS) {
	char *str = PG_GETARG_CSTRING(0);
	Basis *basis = c_basis_in(str);
	PG_RETURN_BASIS_P(basis);
}

Datum basis_out(PG_FUNCTION_ARGS) {
	Basis *basis = PG_GETARG_BASIS_P(0);
	char *str = c_basis_out(basis);
	PG_RETURN_CSTRING(str);
}

Datum new_basis(PG_FUNCTION_ARGS) {
	float4 initval;
	float4 istddev;
	bool hasinit;
	Basis *basis;

	hasinit = !PG_ARGISNULL(0);
	if (hasinit) {
		initval = PG_GETARG_FLOAT4(0);
	} else {
		initval = 0;
	}
	if (fcinfo->nargs > 1) {
		if (PG_ARGISNULL(1)) {
			ereport(ERROR, (errcode(ERRCODE_NULL_VALUE_NOT_ALLOWED),
							errmsg("new_basis() - istddev cannot be NULL")));
		}
		istddev = PG_GETARG_FLOAT4(1);
		if (istddev <= 0) {
			ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
							errmsg("new_basis() - istddev must be positive")));
		}
	} else {
		istddev = 0;
	}

	basis = c_new_basis(initval, istddev, hasinit);
	PG_RETURN_BASIS_P(basis);
}

Datum get_info(PG_FUNCTION_ARGS) {
	TupleDesc tupdesc;
	Basis *basis;
	HeapTupleHeader tuple;
	if (get_call_result_type(fcinfo, NULL, &tupdesc) != TYPEFUNC_COMPOSITE) {
		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
						errmsg("function returning record called in context "
								"that cannot accept type record")));
	}
	tupdesc = BlessTupleDesc(tupdesc);
	basis = PG_GETARG_BASIS_P(0);
	tuple = c_get_info(basis, tupdesc);
	PG_RETURN_HEAPTUPLEHEADER(tuple);
}

Datum expect(PG_FUNCTION_ARGS) {
	float4 value;
	bool isnull;
	Basis *basis = PG_GETARG_BASIS_P(0);
	value = c_expect(&basis->pdf, &isnull);
	if (isnull) {
		PG_RETURN_NULL();
	}
	PG_RETURN_FLOAT4(value);
}

Datum stddev(PG_FUNCTION_ARGS) {
	float4 value;
	bool isnull;
	Basis *basis = PG_GETARG_BASIS_P(0);
	value = c_stddev(&basis->pdf, &isnull);
	if (isnull) {
		PG_RETURN_NULL();
	}
	PG_RETURN_FLOAT4(value);
}

Datum confint(PG_FUNCTION_ARGS) {
	TupleDesc tupdesc;
	Basis *basis;
	float4 mass, minv, maxv;
	Datum values[3];
	bool isnull[3] = {0};
	HeapTuple tuple;
	if (get_call_result_type(fcinfo, NULL, &tupdesc) != TYPEFUNC_COMPOSITE) {
		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
						errmsg("function returning record called in context "
								"that cannot accept type record")));
	}
	tupdesc = BlessTupleDesc(tupdesc);
	basis = PG_GETARG_BASIS_P(0);
	mass = PG_GETARG_FLOAT4(1);
	mass = c_confint(&basis->pdf, mass, &minv, &maxv);
	values[0] = Float4GetDatum(mass);
	if (mass > 0) {
		values[1] = Float4GetDatum(minv);
		values[2] = Float4GetDatum(maxv);
	} else {
		isnull[1] = true;
		isnull[2] = true;
	}
	tuple = heap_form_tuple(tupdesc, values, isnull);
	PG_RETURN_HEAPTUPLEHEADER(tuple->t_data);
}

Datum diverge(PG_FUNCTION_ARGS) {
	Basis *b1 = PG_GETARG_BASIS_P(0);
	Basis *b2 = PG_GETARG_BASIS_P(1);
	float4 value = c_diverge(&b1->pdf, &b2->pdf);
	PG_RETURN_FLOAT4(value);
}
