/*
 * basis.h - representation of the user defined type
 */

#ifndef __BASIS_H__
#define __BASIS_H__

/* Default confidence interval (a non-outlier should be contained). */
#define ERROR_THOLD 0.99

/* Default divergence threshold (pdf can change at most this much). */
#define DELTA_THOLD 0.05

/*----------------------------------------------------------------------------*\
  Internal Structures
\*----------------------------------------------------------------------------*/

/*
 * PDF header (data follows end of struct).
 */
typedef struct {

	/* pdf class; see #define's below */
	char type;

	/* index of model used (-1 if none) */
	int16 model;

	/* variable length of the data array */
	int16 length;

	/* actual parameters, bins, etc. */
	float4 data[0];

	/* if model > -1, append the predictors used */
	/* (i.e., vi/ni and final vn/nn in erace_var) */

} PDF;

/*
 * Empty distribution (initval is null).
 */
#define PDF_ZERO '?'

/*
 * Use the original values (i.e. initval and istddev).
 * This type is only valid when the PDF is contained in a Basis.
 */
#define PDF_INIT 'i'

/*
 * A single (certain) value with probability one.
 */
#define PDF_CERT 'c'

/*
 * Gaussian distribution; data[0] is the mean, data[1] is the stddev.
 */
#define PDF_NORM 'n'

/*----------------------------------------------------------------------------*\
  PDF Support Functions
\*----------------------------------------------------------------------------*/

/* Gets the basis containing the given pdf. */
#define PDF_BASIS(p) ((Basis *)((char *)(p) - sizeof(Basis) + sizeof(PDF)))

float4 c_expect(PDF *pdf, bool *isnull);
float4 c_stddev(PDF *pdf, bool *isnull);
float4 c_confint(PDF *pdf, float4 mass, float4 *out_minv, float4 *out_maxv);
float4 c_diverge(PDF *p1, PDF *p2);

/*----------------------------------------------------------------------------*\
  External Structures
\*----------------------------------------------------------------------------*/

/*
 * Random variable data structure; packages everything necessary for the
 * inference algorithms (instead of passing/returning multiple values).
 */
typedef struct {

	/* varlena header (do not touch directly!) */
	int32 vl_len_;

	/* original value/evidence (if any) */
	float4 initval;

	/* original standard deviation */
	float4 istddev;

	/* true if initval is NOT NULL */
	bool hasinit;

	/* true if initval seems erroneous */
	bool suspect;

	/* round number when pdf/flags last updated */
	int16 round;

	/* current prediction (probability distribution) */
	PDF pdf;

} Basis;

/*
 * Postgres V1 Function Macros
 */
#define DatumGetBasisP(X) ((Basis *) PG_DETOAST_DATUM(X))
#define PG_GETARG_BASIS_P(n) DatumGetBasisP(PG_GETARG_DATUM(n))
#define PG_RETURN_BASIS_P(x) PG_RETURN_POINTER(x)

Basis *update_basis(Basis *basis, PDF *pdf);
Basis *revert_basis(Basis *basis);

#define NORM_BUF_SIZE sizeof(PDF) + 8
PDF *get_init_pdf(Basis *basis, char *buf);

#endif /* __BASIS_H__ */
