package src;

import java.util.Random;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Simulation parameters and demographic distributions.
 */
class Model {

  /**
   * Single entry in a cumulative probability distribution.
   */
  private static class Range {
    int minAge;
    int width;
    float cprob;

    public Range(Element node) {
      minAge = Integer.parseInt(node.getAttribute("minAge"));
      int maxAge = Integer.parseInt(node.getAttribute("maxAge"));
      width = maxAge - minAge + 1;
      cprob = Float.parseFloat(node.getAttribute("prob"));
    }
  } // private static class Range

  /**
   * Cumulative probability distribution (function of year and age).
   */
  private static class CDist {
    int minYear;
    int maxYear;
    Range[] bins;

    public CDist(Element node) {
      minYear = att2int(node, "minYear", Integer.MIN_VALUE);
      maxYear = att2int(node, "maxYear", Integer.MAX_VALUE);

      NodeList list = node.getElementsByTagName("Range");
      int cnt = list.getLength();
      bins = new Range[cnt];
      for (int i = 0; i < cnt; i++) {
        bins[i] = new Range((Element) list.item(i));
      }
      for (int i = 1; i < cnt; i++) {
        // make the probabilities cumulative
        bins[i].cprob += bins[i - 1].cprob;
      }
      if (bins[cnt - 1].cprob != 1) {
        throw new IllegalArgumentException("invalid distribution for " + node.getTagName());
      }
    }
  } // private static class CDist

  /**
   * Parses an integer from the given attribute; returns def if not found.
   */
  private static int att2int(Element node, String att, int def) {
    String str = node.getAttribute(att);
    if (str.length() == 0) {
      return def;
    }
    return Integer.parseInt(str);
  }

  /**
   * Poisson distribution for a given time period.
   */
  private static class PDist {
    int minYear;
    int maxYear;
    float lamda;

    public PDist(Element node) {
      minYear = Integer.parseInt(node.getAttribute("minYear"));
      maxYear = Integer.parseInt(node.getAttribute("maxYear"));
      lamda = Float.parseFloat(node.getAttribute("lamda"));
    }
  } // private static class PDist

  // --------------------------------------------------------------------------

  /** Random number generator shared by all methods. */
  public static final Random RNG;

  public static final int INIT_CNT;
  public static final int MIN_YEAR;
  public static final int MAX_YEAR;
  public static final int MAX_PREG;
  public static final float IMG_RATE;
  public static final float DIV_RATE;

  private static final Range[] founderAge;
  private static final Range[] migrantAge;
  private static final CDist[] marryAges;
  private static final PDist[] birthRates;
  private static final CDist[] deathAges;

  // --------------------------------------------------------------------------

  /**
   * Constructs parameters from the current values in the XML file.
   */
  static {

    // parse the XML config file
    Document doc = null;
    try {
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      DocumentBuilder builder = factory.newDocumentBuilder();
      doc = builder.parse("params.xml");
    } catch (Exception exc) {
      exc.printStackTrace();
      System.exit(1);
    }

    // initialize fields from XML
    Element root = doc.getDocumentElement();
    RNG = new Random(Integer.parseInt(root.getAttribute("rng_seed")));
    INIT_CNT = Integer.parseInt(root.getAttribute("init_cnt"));
    MIN_YEAR = Integer.parseInt(root.getAttribute("min_year"));
    MAX_YEAR = Integer.parseInt(root.getAttribute("max_year"));
    MAX_PREG = Integer.parseInt(root.getAttribute("max_preg"));
    IMG_RATE = Float.parseFloat(root.getAttribute("img_rate"));
    DIV_RATE = Float.parseFloat(root.getAttribute("div_rate"));

    // construct distributions based on age only
    founderAge = new1D(root, "FounderAge");
    migrantAge = new1D(root, "MigrantAge");
    marryAges = new2D(root, "MarryAge");
    birthRates = newPD(root, "BirthRate");
    deathAges = new2D(root, "DeathAge");

  } // static

  /**
   * Constructs a distribution based on age only.
   */
  private static Range[] new1D(Element node, String name) {
    NodeList list = node.getElementsByTagName(name);
    return new CDist((Element) list.item(0)).bins;
  }

  /**
   * Constructs a distribution based on year and age.
   */
  private static CDist[] new2D(Element node, String name) {
    NodeList list = node.getElementsByTagName(name);
    int cnt = list.getLength();
    CDist[] ret = new CDist[cnt];
    for (int i = 0; i < cnt; i++) {
      ret[i] = new CDist((Element) list.item(i));
    }
    return ret;
  }

  /**
   * Constructs a poisson distribution based on year.
   */
  private static PDist[] newPD(Element node, String name) {
    NodeList list = node.getElementsByTagName(name);
    int cnt = list.getLength();
    PDist[] ret = new PDist[cnt];
    for (int i = 0; i < cnt; i++) {
      ret[i] = new PDist((Element) list.item(i));
    }
    return ret;
  }

  // --------------------------------------------------------------------------

  public static int founderAge() {
    return randomAge(founderAge);
  }

  public static int migrantAge() {
    return randomAge(migrantAge);
  }

  public static int marryAge(int year) {
    return randomAge(marryAges, year);
  }

  public static int deathAge(int year) {
    return randomAge(deathAges, year);
  }

  /**
   * Samples a random age from the given distribution.
   */
  private static int randomAge(Range[] bins) {
    float rand = RNG.nextFloat();
    for (int i = 0; i < bins.length; i++) {
      if (rand < bins[i].cprob) {
        return RNG.nextInt(bins[i].width) + bins[i].minAge;
      }
    }
    throw new IllegalStateException("wrongly assumed unreachable code");
  }

  /**
   * Samples a random age from the distribution for the given year.
   */
  private static int randomAge(CDist[] dists, int year) {
    for (int i = 0; i < dists.length; i++) {
      CDist dist = dists[i];
      if (year >= dist.minYear && year <= dist.maxYear) {
        return randomAge(dist.bins);
      }
    }
    throw new IllegalArgumentException("no distribution for the year " + year);
  }

  /**
   * Returns the birth year of the calling individual's next child.
   */
  public static int nextChild(int year) {
    for (int i = 0; i < birthRates.length; i++) {
      PDist dist = birthRates[i];
      if (year >= dist.minYear && year <= dist.maxYear) {
        // see "generating exponential variates" in Wikipedia
        double exp = -1 * Math.log(RNG.nextDouble()) / dist.lamda;
        return (year + (int) Math.round(exp));
      }
    }
    throw new IllegalArgumentException("no distribution for the year " + year);
  }

} // class Model
