package src;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;

/**
 * Simulated individual and supporting data structures.
 */
class Person implements Comparable<Person> {

  static int next_id = 1000001 - Model.INIT_CNT;

  static enum Type {
    FOUNDER, MIGRANT, NEWBORN
  }

  static enum Gender {
    FEMALE, MALE
  }

  static enum Status {
    SINGLE, MARRIED, DIVORCED, WIDOWED
  }

  // --------------------------------------------------------------------------

  // set in constructor
  int ind_id;
  Type type;
  Gender gender;
  Status status;
  int birth_year;
  int death_year;
  int marry_ready;

  // set externally
  int marry_first;
  int next_child;
  Person father;
  Person mother;
  ArrayList<Person> spouses;
  ArrayList<Person> children;
  String role;

  // --------------------------------------------------------------------------

  /**
   * Constructs a new person, given the current year.
   */
  public Person(int year, Type type) {

    // initialize common fields
    ind_id = next_id++;
    this.type = type;
    gender = Model.RNG.nextBoolean() ? Gender.FEMALE : Gender.MALE;
    status = Status.SINGLE;

    // generate event years
    switch (type) {
      case FOUNDER:
        birth_year = year - Model.founderAge();
        break;
      case MIGRANT:
        birth_year = year - Model.migrantAge();
        break;
      case NEWBORN:
        birth_year = year;
        break;
    }
    death_year = birth_year + Model.deathAge(birth_year);
    marry_ready = birth_year + Model.marryAge(birth_year);

    // initialize collections
    spouses = new ArrayList<Person>();
    children = new ArrayList<Person>();

  } // public Person(int year, Type type)

  /**
   * Updates the marriage status of this person.
   */
  public void marry(Person spouse, int year) {
    status = Status.MARRIED;
    if (marry_first == 0) {
      marry_first = year;
    }
    spouses.add(0, spouse);
  }

  /**
   * Updates the marriage status of both spouses.
   */
  public void split(Status status) {
    this.status = status;
    this.marry_ready = 9999;
    Person spouse = spouses.get(0);
    spouse.status = status;
    spouse.marry_ready = 9999;
  }

  /**
   * Updates and returns child bearing status.
   */
  public boolean fertile(int year) {

    // first check qualifications
    if (gender != Gender.FEMALE) {
      return false;
    }
    if (year - birth_year > Model.MAX_PREG) {
      return false;
    }
    if (status != Status.MARRIED) {
      return false;
    }

    // schedule the next child
    if (next_child < year) {
      next_child = Model.nextChild(year);
    }
    return (next_child == year);

  } // public boolean fertile(int year)

  /**
   * Adds the newborn child to this mother and her spouse.
   */
  public void birth(Person child) {
    child.mother = this;
    children.add(child);
    Person spouse = spouses.get(0);
    child.father = spouse;
    spouse.children.add(child);
  }

  /**
   * Is the given person a close relative?
   */
  public boolean related(Person rel) {
    return (rel == father || rel == mother || children.contains(rel));
  }

  // --------------------------------------------------------------------------

  /**
   * Should this person be included in the final output?
   */
  public boolean include() {

    // discard isolated individuals (no parents/children)
    if (father == null && mother == null && children.size() == 0) {
      return false;
    }

    // discard founders since we're not simulating their
    // marriages and children before the initial year
    if (type == Type.FOUNDER) {
      return false;
    }

    // include by default
    return true;

  } // public boolean include()

  /**
   * Compares this object with the specified object for order.
   */
  public int compareTo(Person that) {
    return (this.ind_id - that.ind_id);
  }

  /**
   * Returns a string representation of the Person.
   */
  public String toString() {
    StringBuilder str = new StringBuilder();
    str.append(ind_id);
    str.append('\t');
    str.append(type);
    str.append('\t');
    str.append(gender);
    str.append('\t');
    str.append(status);
    str.append('\t');
    str.append(birth_year);
    str.append('-');
    str.append(death_year);
    str.append('\t');
    str.append("m~");
    str.append(marry_ready);
    str.append('\t');
    str.append("m=");
    str.append(marry_first);
    str.append('\t');
    str.append("c=");
    str.append(children.size());
    return str.toString();
  }

  /**
   * Writes the person tuple to the given file.
   */
  public void printInd(PrintWriter out) {

    out.print(ind_id);
    out.print(',');

    out.print(birth_year);
    out.print(',');

    if (marry_first > 0) {
      out.print(marry_first);
    } else {
      // out.print("\\N");
    }
    out.print(',');

    out.print(death_year);
    out.print(',');

    // for (int i = 0; i < 3; i++) {
    // out.print("TRUE,");
    // }

    out.print(children.size() > 0 ? "TRUE" : "FALSE");
    out.print(',');
    out.print(gender == Gender.FEMALE ? "TRUE" : "FALSE");
    out.print(',');
    out.println(Model.RNG.nextFloat());

  } // public void printInd(PrintWriter out)

  /**
   * Writes the relative tuples to the given file.
   */
  public void printRels(PrintWriter out) {

    // build a list of relatives to include
    ArrayList<Person> rels = new ArrayList<Person>();
    if (father != null && father.include()) {
      father.role = "P:parent";
      rels.add(father);
    }
    if (mother != null && mother.include()) {
      mother.role = "P:parent";
      rels.add(mother);
    }
    for (Person spouse : spouses) {
      if (spouse.include()) {
        spouse.role = "S:spouse";
        rels.add(spouse);
      }
    }
    for (Person child : children) {
      if (child.include()) {
        child.role = "C:child";
        rels.add(child);
      }
    }

    // print relatives in ascending order
    Collections.sort(rels);
    for (Person rel : rels) {
      out.print(ind_id);
      out.print(',');
      out.print(rel.ind_id);
      out.print(',');
      out.println(rel.role);
    }

  } // public void printRels(PrintWriter out)

} // class Person implements Comparable<Person>
