package src;

/*
 * Related approaches:
 * 
 * B. Elliott, et al. Efficient Evaluation of Inbreeding Queries on Pedigree
 * Data. In Proc of Scientific and Statistical Database Management, 2007.
 * 
 * V. Ollikainen. Simulation Techniques for Disease Gene Localization in
 * Isolated Populations. Ph.D. thesis, University of Helsinki, 2002.
 * 
 * H. Toivonen, et al. Data Mining Applied to Linkage Disequilibrium Mapping.
 * American Society of Human Genetics, 67:133-145, 2000. (see Appendix B)
 */

import src.Person.Gender;
import src.Person.Status;
import src.Person.Type;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.PriorityQueue;

/**
 * Ad hoc population generator (i.e. no target population size distribution).
 */
public class PopGen {

  /** Current population of living individuals. */
  private LinkedList<Person> m_pop;

  /** Buffer of subsequent individuals to output. */
  private PriorityQueue<Person> m_out;

  /** The ind_id of the next person to output. */
  private int m_next;

  /** Output file for the person table. */
  private PrintWriter m_ptab;

  /** Output file for the relative table. */
  private PrintWriter m_rtab;

  /** Output file for the stats table. */
  private PrintWriter m_stab;

  // --------------------------------------------------------------------------

  /**
   * Application entry point.
   */
  public static void main(String[] args) throws IOException {
    PopGen pg = new PopGen();
    pg.simulate();
    pg.close();
  }

  /**
   * Default constructor.
   */
  public PopGen() throws IOException {
    m_pop = new LinkedList<Person>();
    m_out = new PriorityQueue<Person>();
    m_next = Person.next_id;
    m_ptab = new PrintWriter("out/person.csv");
    m_rtab = new PrintWriter("out/relative.csv");
    m_stab = new PrintWriter("out/stats.csv");
    m_ptab.println("ind_id,birth,marry,death,parent,female,random");
    m_rtab.println("ind_id,rel_id,role");
    m_stab.println("year,immg,death,sep,mar,birth,pop,fams");
  }

  /**
   * Closes the output files.
   */
  private void close() {
    m_ptab.close();
    m_rtab.close();
    m_stab.close();
  }

  // --------------------------------------------------------------------------

  /**
   * Simulation entry point.
   */
  private void simulate() {

    // create the initial population
    for (int i = 0; i < Model.INIT_CNT; i++) {
      m_pop.add(new Person(Model.MIN_YEAR, Type.FOUNDER));
    }

    // simulate events for each year
    for (int year = Model.MIN_YEAR; year <= Model.MAX_YEAR; year++) {

      // add immigrants to the population
      int icnt = Math.round(Model.IMG_RATE * m_pop.size());
      for (int i = 0; i < icnt; i++) {
        m_pop.add(new Person(year, Type.MIGRANT));
      }

      // remove and output deceased people
      int dcnt = 0;
      ListIterator<Person> iter = m_pop.listIterator();
      while (iter.hasNext()) {
        Person ind = iter.next();
        if (ind.death_year <= year) {
          iter.remove();
          output(ind);
          if (ind.status == Status.MARRIED) {
            ind.split(Status.WIDOWED);
          }
          dcnt++;
        }
      }

      // separate the unlucky random couples
      ArrayList<Person> wives = new ArrayList<Person>();
      for (Person ind : m_pop) {
        if (ind.gender == Gender.FEMALE && ind.status == Status.MARRIED) {
          wives.add(ind);
        }
      }
      int scnt = Math.round(Model.DIV_RATE * wives.size());
      Collections.shuffle(wives, Model.RNG);
      for (int i = 0; i < scnt; i++) {
        wives.get(i).split(Status.DIVORCED);
      }

      // identify marriage candidates
      ArrayList<Person> gals = new ArrayList<Person>();
      ArrayList<Person> guys = new ArrayList<Person>();
      for (Person ind : m_pop) {
        if (ind.status != Status.MARRIED && ind.marry_ready <= year) {
          if (ind.gender == Gender.FEMALE) {
            gals.add(ind);
          } else {
            guys.add(ind);
          }
        }
      }

      // arrange new marriages
      int mcnt = Math.min(gals.size(), guys.size());
      Collections.shuffle(gals, Model.RNG);
      Collections.shuffle(guys, Model.RNG);
      for (int i = 0; i < mcnt; i++) {
        Person gal = gals.get(i);
        Person guy = guys.get(i);
        if (!gal.related(guy)) {
          gal.marry(guy, year);
          guy.marry(gal, year);
        }
      }

      // add births to the population
      ArrayList<Person> moms = new ArrayList<Person>();
      for (Person ind : m_pop) {
        if (ind.fertile(year)) {
          moms.add(ind);
        }
      }
      int bcnt = 0;
      for (Person mom : moms) {
        Person child = new Person(year, Type.NEWBORN);
        mom.birth(child);
        m_pop.add(child);
        bcnt++;
      }

      // count how many couples can still have children
      int fcnt = 0;
      for (Person ind : m_pop) {
        if (ind.gender == Gender.FEMALE)
          if (year - ind.birth_year <= Model.MAX_PREG)
            if (ind.status == Status.MARRIED)
              fcnt++;
      }

      // show statistics
      if (year % 25 == 0) {
        System.out.println("Year = " + year);
      }
      m_stab.print(year + ",");
      m_stab.print(icnt + ",");
      m_stab.print(dcnt + ",");
      m_stab.print(scnt + ",");
      m_stab.print(mcnt + ",");
      m_stab.print(bcnt + ",");
      m_stab.print(m_pop.size() + ",");
      m_stab.println(fcnt);

    } // for

    // output the remaining individuals
    for (Person ind : m_pop) {
      output(ind);
    }

  } // private void simulate()

  /**
   * Outputs tuples corresponding to the given person.
   */
  private void output(Person ind) {

    // output individuals in order
    m_out.add(ind);
    while (m_out.isEmpty() == false && m_out.element().ind_id == m_next) {

      // get the next person
      ind = m_out.remove();
      m_next++;

      // output sub
      if (ind.include()) {
        ind.printInd(m_ptab);
        ind.printRels(m_rtab);
      }

    } // while

  } // private void output(Person ind)

} // public class PopGen
