-- original data
DROP TABLE IF EXISTS locs;
DROP TABLE IF EXISTS data;
DROP TABLE IF EXISTS conn;

-- derived data
DROP TABLE IF EXISTS near;
DROP TABLE IF EXISTS clean;

--------------------------------------------------------------------------------

CREATE TABLE locs (
  moteid integer NOT NULL,
  xloc real NOT NULL,
  yloc real NOT NULL
);

COMMENT ON TABLE locs IS
'locations in meters relative to the upper right corner';

\copy locs FROM 'src/mote_locs.txt' DELIMITER ' ' NULL ''

ANALYZE VERBOSE locs;
SELECT count(*) FROM locs; -- 54

--------------------------------------------------------------------------------

CREATE TABLE data0 (
  date text NOT NULL,
  time text NOT NULL,
  epoch integer NOT NULL,
  moteid integer,
  temperature real,
  humidity real,
  light real,
  voltage real
);

-- 2,313,682 rows in 32,459 ms
\copy data0 FROM 'src/data.txt' DELIMITER ' ' NULL ''

-- 2,313,359 rows in 87,293 ms
CREATE TABLE data AS
  -- aggregate multiple readings during the same second
  -- (can use min/max/avg as alternatvies to array_agg)
  SELECT
    (array_agg(ts))[1] AS ts,
    (array_agg(epoch))[1] AS epoch,
    second,
    moteid,
    (array_agg(temperature))[1] AS temperature,
    (array_agg(humidity))[1] AS humidity,
    (array_agg(light))[1] AS light,
    (array_agg(voltage))[1] AS voltage
  FROM (
    SELECT *, -- convert timestamps into elapsed seconds
      date_part('epoch', ts - '2004-02-28 00:58:45')::int AS second
    FROM (
      SELECT *, -- convert dates/times into timestamps
        (date || ' ' || time)::timestamp AS ts
      FROM data0
    ) AS q1
  ) AS q2
  GROUP BY second, moteid
  ORDER BY second, moteid;

DROP TABLE data0;

ALTER TABLE data
  ALTER ts SET NOT NULL,
  ALTER epoch SET NOT NULL,
  ALTER second SET NOT NULL;

COMMENT ON TABLE data IS
'readings from 54 sensors about every 31 seconds';

-- 32,684 ms
CREATE INDEX data__second ON data (second);
CREATE INDEX data__moteid ON data (moteid);
CREATE INDEX data__second_moteid ON data (second, moteid);
CREATE INDEX data__moteid_second ON data (moteid, second);

ANALYZE VERBOSE data;

--------------------------------------------------------------------------------

CREATE TABLE conn (
  sender integer NOT NULL,
  receiver integer NOT NULL,
  probability real
);

COMMENT ON TABLE conn IS
'aggregate connectivity data averaged over all time';

\copy conn FROM 'src/connectivity.txt' DELIMITER ' ' NULL ''

ANALYZE VERBOSE conn;
SELECT count(*) FROM conn; -- 2,971

--------------------------------------------------------------------------------

CREATE TABLE near (
  id1 integer NOT NULL,
  id2 integer NOT NULL,
  dist real NOT NULL
);

COMMENT ON TABLE near IS
'which motes are near each other (in the same room)';

INSERT INTO near
  SELECT *
  FROM (
    SELECT a.moteid AS id1, b.moteid AS id2,
      point(a.xloc, a.yloc) <-> point(b.xloc, b.yloc) AS dist
    FROM locs AS a
      INNER JOIN locs AS b ON a.moteid <> b.moteid -- symetric
  ) AS q1
  WHERE dist < 6 -- meters
  ORDER BY id1, id2;

-- remove any edges that go through walls
DELETE FROM near
WHERE (id1 =  8 AND id2 = 53)
   OR (id1 =  8 AND id2 = 54)
   OR (id1 =  9 AND id2 = 54)
   OR (id1 =  9 AND id2 = 11)
   OR (id1 = 10 AND id2 = 11)
   OR (id1 = 28 AND id2 = 31)
   OR (id1 = 29 AND id2 = 31)
   OR (id1 = 48 AND id2 = 52)
   -- other direction
   OR (id2 =  8 AND id1 = 53)
   OR (id2 =  8 AND id1 = 54)
   OR (id2 =  9 AND id1 = 54)
   OR (id2 =  9 AND id1 = 11)
   OR (id2 = 10 AND id1 = 11)
   OR (id2 = 28 AND id1 = 31)
   OR (id2 = 29 AND id1 = 31)
   OR (id2 = 48 AND id1 = 52);

-- can't run VACUUM in a script
CREATE INDEX near__id1_id2 ON near (id1, id2);
CLUSTER near USING near__id1_id2;
DROP INDEX near__id1_id2;

ANALYZE VERBOSE near;
SELECT count(*) FROM near; -- 160

--------------------------------------------------------------------------------

CREATE TABLE clean0 (
  date text,
  time text,
  epoch integer,
  moteid integer,
  temperature real,
  humidity real,
  light real,
  voltage real
);

-- 2,269,503 rows in 31,254 ms
\copy clean0 FROM 'out/clean.txt' DELIMITER ' ' NULL ''

-- 33,896 ms
CREATE TABLE clean AS
  SELECT ts, second, moteid, -- epoch column is inconsistent
    temperature, humidity, light, voltage
  FROM (
    SELECT *, -- convert timestamps into elapsed seconds
      date_part('epoch', ts - '2004-02-28 00:58:45')::int AS second
    FROM (
      SELECT *, -- convert dates/times into timestamps
        (date || ' ' || time)::timestamp AS ts
      FROM clean0
    ) AS q1
  ) AS q2
  ORDER BY second, moteid;

DROP TABLE clean0;

ALTER TABLE clean
  ALTER ts SET NOT NULL,
  ALTER second SET NOT NULL,
  ALTER moteid SET NOT NULL;

COMMENT ON TABLE clean IS
'cleaned up version of data (via lower/upper bounds)';

-- 30,702 ms
CREATE INDEX clean__second ON clean (second);
CREATE INDEX clean__moteid ON clean (moteid);
CREATE INDEX clean__second_moteid ON clean (second, moteid);
CREATE INDEX clean__moteid_second ON clean (moteid, second);

ANALYZE VERBOSE clean;
