"""
Test driver and monitor for the experiments.
"""

from psycopg2 import connect
from time import ctime
from time import time
import sys

def usage():
    print "Usage: python -u exper.py <DATABASE> " \
        "<VERSION> <m|c> <ROUNDS> [TABLESPACE]"
    sys.exit(1)

def tab_name(m, c, r):
    # v = version, m = missing, c = corrupt, r = round
    return "v{0}_m{1:02}_c{2:02}_r{3}".format(VERSION, m, c, r)

def print_tt(t0):
    tt = time() - t0
    print "-- {0:.3f} sec\n".format(tt)

def print_hr():
    print "------------------------------------------------------------\n"

def execute(sql, stats=False):
    """Utility function; prints and executes the given sql."""
    print sql
    t0 = time()
    curs.execute(sql)
    if stats:
        print_tt(t0)

def round_0(m, c):
    """Execute initialization queries for round 0."""
    name = tab_name(m, c, 0)
    if name in tabs:
        return
    execute("""CREATE OR REPLACE VIEW params AS
        SELECT 0.{0:02} AS missing, 0.{1:02} AS corrupt;
        """.format(m, c))
    execute("""CREATE TABLE {0} TABLESPACE {1} AS
        SELECT * FROM input;
        """.format(name, TABLESPACE), True)
    conn.commit()
    execute("""CREATE UNIQUE INDEX {0}__nid
        ON {0} (nid) TABLESPACE {1};
        """.format(name, TABLESPACE), True)
    if DBNAME == "sensor":
        onehop(name)
    execute("ANALYZE " + name + ";\n")
    conn.commit()
    print_hr()
    result_ins(m, c, 0, name)

def onehop(name):
    """Clears out half of the one-hop neighbors of missing nodes."""
    execute("SELECT * FROM node LIMIT 0;")
    cols = [tup[0] for tup in curs.description]
    rest = ""
    for c in cols[2:]:
        rest = rest + "   ,{0} = new_basis(NULL)\n        ".format(c)
    sql = """UPDATE {0}
        SET {1} = new_basis(NULL)
        {2}WHERE random() > 0.5 AND nid IN (
          SELECT DISTINCT n.nid
          FROM {0} AS i
            INNER JOIN link AS l ON i.nid = l.dst
            INNER JOIN {0} AS n ON l.src = n.nid
          WHERE expect(i.{1}) IS NULL
            AND expect(n.{1}) IS NOT NULL);
        """.format(name, cols[1], rest)
    execute(sql, True)

def round_r(m, c, r):
    """Execute inference queries for the given round."""
    name = tab_name(m, c, r)
    if name in tabs:
        return
    prev = tab_name(m, c, r - 1)
    if DBNAME == 'popgen':
        if r % 2 == 1:
            execute("SELECT model_sync('p');")
            link = "link_p"
        else:
            execute("SELECT model_sync('c');")
            link = "link_c"
    else:
        link = "link"
    execute("SELECT set_round(" + str(r) + ");\n")
    execute("""CREATE CAST ({0} AS node)
        WITHOUT FUNCTION AS IMPLICIT;
        """.format(prev))
    execute("""CREATE TABLE {0} TABLESPACE {1} AS
        SELECT (erace(i, n)).*
        FROM {2} AS i
          LEFT JOIN {3} AS l ON i.nid = l.dst
          LEFT JOIN {2} AS n ON l.src = n.nid
        GROUP BY i;
        """.format(name, TABLESPACE, prev, link), True)
    execute("""CREATE UNIQUE INDEX {0}__nid
        ON {0} (nid) TABLESPACE {1};
        """.format(name, TABLESPACE), True)
    execute("ANALYZE " + name + ";\n")
    conn.commit()
    print_hr()
    result_ins(m, c, r, name)

def iterate():
    """Main algorithm; runs inference on each (m,c,r)."""
    result_cre()
    t0 = time()
    for m in M_RANGE:
        for c in C_RANGE:
            round_0(m, c)
            for r in R_RANGE:
                round_r(m, c, r)
    tt = time() - t0
    print "-- iterate: {0:.0f} min {1:.0f} sec\n".format(tt / 60, tt % 60)

def result_cre():
    """Creates the result tables (as needed)."""
    for rtab in RESULTS:
        if rtab in tabs:
            continue
        execute("""CREATE TABLE eracer.{0} AS
            SELECT *,
              NULL::timestamp AS ts,
              NULL::integer AS v,
              NULL::integer AS m,
              NULL::integer AS c,
              NULL::integer AS r
            FROM {0} LIMIT 0;
            """.format(rtab))
    conn.commit()
    print_hr()

def result_ins(m, c, r, name):
    """Computes and inserts this round's results."""
    execute("""CREATE OR REPLACE VIEW current AS
        SELECT * FROM {0};
        """.format(name))
    # special case: no results at round 0, just status
    if r == 0:
        todo = RESULTS[:1]
    else:
        todo = RESULTS
    for rtab in todo:
        execute("""INSERT INTO eracer.{0}
            SELECT *, now(), {1}, {2}, {3}, {4}
            FROM {0};
            """.format(rtab, VERSION, m, c, r), True)
    conn.commit()
    print_hr()

if __name__ == "__main__":
    """Parses arguments and initializes globals (conn, curs, tabs)"""
    if len(sys.argv) < 5:
        usage()
    DBNAME = sys.argv[1]
    VERSION = sys.argv[2]
    if sys.argv[3] == 'm':
        if DBNAME == "sensor":
            M_RANGE = [         4, 10, 18, 28, 40, 58, 79]
            C_RANGE = [0]
        else:
            M_RANGE = [10, 20, 30, 40, 50, 60, 70, 80, 90]
            C_RANGE = [0]
    elif sys.argv[3] == 'c':
        if DBNAME == "sensor":
            M_RANGE = [18] # about 50%
            C_RANGE = [ 8, 17, 26, 36, 45, 56, 68, 83    ]
        else:
            M_RANGE = [30]
            C_RANGE = [ 7, 14, 21, 28, 35, 42, 49, 56, 63]
    else:
        usage()
    R_RANGE = range(1, int(sys.argv[4]) + 1)
    if len(sys.argv) > 5:
        TABLESPACE = sys.argv[5]
    else:
        TABLESPACE = "pg_default"
    RESULTS = ["status", "summary", "res_aq", "res_dc"]
    # initialize other globals
    print_hr()
    conn = connect("dbname=" + DBNAME)
    curs = conn.cursor()
    print "-- start time:\n-- {0}\n".format(ctime())
    execute("SET search_path TO v{0}, eracer;\n".format(VERSION))
    execute("SELECT set_version(" + VERSION + ");\n")
    # get all table names to prevent redoing any work
    execute("""SELECT tablename FROM pg_tables
        WHERE schemaname IN ('v{0}', 'eracer');
        """.format(VERSION))
    try:
        tabs = curs.fetchall()
        tabs = dict([(tabs[i][0], i) for i in range(len(tabs))])
    except:
        tabs = []
    # call the main function: iterate or cleanup
    iterate()
    print "-- end time:\n-- {0}\n".format(ctime())
