--
-- example session using the genealogy data
-- (what to do after running install.sql)
--

SET search_path TO eracer, public;

--------------------------------------------------------------------------------
-- STEP 1: construct graphical model

ALTER TABLE node
  -- inferred values
  ADD COLUMN b basis NOT NULL,
  ADD COLUMN m basis NOT NULL,
  ADD COLUMN d basis NOT NULL;

-- 5,000,000 rows in 29,708 ms
INSERT INTO node
  SELECT
    ind_id,
    new_basis(birth),
    new_basis(marry),
    new_basis(death)
  FROM person;

-- 16,163,706 rows in 18,178 ms
INSERT INTO link
  SELECT
    ind_id AS dst,
    rel_id AS src,
    CASE role
      WHEN 'P:parent' THEN 1
      WHEN 'C:child'  THEN 2
      WHEN 'S:spouse' THEN 3
    END AS typ
  FROM relative;

/*
-- add "siblings" to the model
INSERT INTO link
  SELECT DISTINCT
    r1.ind_id, r2.ind_id, 4 AS typ
  FROM relative r1, relative r2
  WHERE r1.role = 'P:parent'
    AND r2.role = 'P:parent'
    AND r1.rel_id = r2.rel_id -- same parent
    AND r1.ind_id != r2.ind_id; -- different child
*/

-- create indexes (see tables.sql)
-- node: 31,642 ms  link: 129,365 ms

--------------------------------------------------------------------------------
-- STEP 2: learn regression coefficients

ALTER TABLE model
  -- intra-tuple coefficients
  ADD COLUMN bi real NOT NULL,
  ADD COLUMN mi real NOT NULL,
  ADD COLUMN di real NOT NULL,
  -- parent coefficients
  ADD COLUMN bp real NOT NULL,
  ADD COLUMN mp real NOT NULL,
  ADD COLUMN dp real NOT NULL,
  -- child coefficients
  ADD COLUMN bc real NOT NULL,
  ADD COLUMN mc real NOT NULL,
  ADD COLUMN dc real NOT NULL,
  -- spouse coefficients
  ADD COLUMN bs real NOT NULL,
  ADD COLUMN ms real NOT NULL,
  ADD COLUMN ds real NOT NULL;

-- 100,000 rows in  44,408 ms (with subquery)
-- 200,000 rows in 122,948 ms (with coalesce)
CREATE TABLE learn AS
  SELECT
    -- individual instances
    min(bi)::real AS bi, min(mi)::real AS mi, min(di)::real AS di,
    -- average neighbor values
    avg(bp)::real AS bp, avg(mp)::real AS mp, avg(dp)::real AS dp,
    avg(bc)::real AS bc, avg(mc)::real AS mc, avg(dc)::real AS dc,
    avg(bs)::real AS bs, avg(ms)::real AS ms, avg(ds)::real AS ds
  FROM (
    SELECT i.nid,
      expect(i.b) AS bi,
      expect(i.m) AS mi,
      expect(i.d) AS di,
      -- parent
      CASE WHEN l.typ = 1 THEN expect(n.b) ELSE NULL END AS bp,
      CASE WHEN l.typ = 1 THEN expect(n.m) ELSE NULL END AS mp,
      CASE WHEN l.typ = 1 THEN expect(n.d) ELSE NULL END AS dp,
      -- child
      CASE WHEN l.typ = 2 THEN expect(n.b) ELSE NULL END AS bc,
      CASE WHEN l.typ = 2 THEN expect(n.m) ELSE NULL END AS mc,
      CASE WHEN l.typ = 2 THEN expect(n.d) ELSE NULL END AS dc,
      -- spouse
      CASE WHEN l.typ = 3 THEN expect(n.b) ELSE NULL END AS bs,
      CASE WHEN l.typ = 3 THEN expect(n.m) ELSE NULL END AS ms,
      CASE WHEN l.typ = 3 THEN expect(n.d) ELSE NULL END AS ds
    FROM node AS i
      LEFT JOIN link AS l ON i.nid = l.dst
      LEFT JOIN node AS n ON l.src = n.nid
  ) AS sub
  -- filter completely missing rows
  WHERE NOT (bi IS NULL AND mi IS NULL AND di IS NULL)
    -- apply domain constraints
    AND COALESCE(di - bi, 0) >= 0 AND COALESCE(di - bi, 0) <= 99
    AND COALESCE(dp - bp, 0) >= 0 AND COALESCE(dp - bp, 0) <= 99
    AND COALESCE(dc - bc, 0) >= 0 AND COALESCE(dc - bc, 0) <= 99
    AND COALESCE(ds - bs, 0) >= 0 AND COALESCE(ds - bs, 0) <= 99
  GROUP BY nid
  ORDER BY random()
  LIMIT 200000;

ANALYZE VERBOSE learn;

--
-- do the one-time regression analysis offline; see learn/README.txt
-- (export learn table, run R and python scripts, import model table)
--

ANALYZE VERBOSE model;
SELECT model_sync();

--------------------------------------------------------------------------------
-- STEP 3: iterative approximate inference (see exper.py)
