--
-- views for generating experiments and evaluating results
--

CREATE SCHEMA v1;
SET search_path TO v1;

CREATE OR REPLACE VIEW params AS
  SELECT 0.00 AS missing, 0.00 AS corrupt;

    COMMENT ON VIEW params
    IS 'missing/corrupt percentages for input view';

CREATE OR REPLACE VIEW current AS
  SELECT * FROM eracer.node;

    COMMENT ON VIEW current
    IS 'alias for node table after current round';

--------------------------------------------------------------------------------

CREATE OR REPLACE VIEW input AS
  SELECT nid,
    -- assume that missing values co-occur in real data
    CASE WHEN missing THEN eracer.new_basis(NULL)
         WHEN corrupt THEN eracer.new_basis((1500 + random() * 399)::real)
         ELSE b END AS b,
    CASE WHEN missing THEN eracer.new_basis(NULL)
         WHEN corrupt THEN eracer.new_basis((1500 + random() * 399)::real)
         ELSE m END AS m,
    CASE WHEN missing THEN eracer.new_basis(NULL)
         WHEN corrupt THEN eracer.new_basis((1500 + random() * 399)::real)
         ELSE d END AS d
  FROM (
    SELECT node2.*,
      random <= missing AS missing,
      random > missing AND random <= missing + corrupt AS corrupt
    FROM params,
      (SELECT *, random() FROM eracer.node) AS node2
  ) AS assign;

COMMENT ON VIEW input
IS 'generates random missing/corrupt data';

--------------------------------------------------------------------------------

CREATE OR REPLACE VIEW output AS
  SELECT n.nid,
    o.birth AS b_orig,
      (bi).initval AS b_init,
      (bi).hasinit AS b_has,
      (bi).suspect AS b_susp,
      (bi).round AS b_rnd,
      (bi).expect AS b_exp,
      (bi).stddev AS b_std,
    o.marry AS m_orig,
      (mi).initval AS m_init,
      (mi).hasinit AS m_has,
      (mi).suspect AS m_susp,
      (mi).round AS m_rnd,
      (mi).expect AS m_exp,
      (mi).stddev AS m_std,
    o.death AS d_orig,
      (di).initval AS d_init,
      (di).hasinit AS d_has,
      (di).suspect AS d_susp,
      (di).round AS d_rnd,
      (di).expect AS d_exp,
      (di).stddev AS d_std
  FROM public.person AS o
    INNER JOIN (
      SELECT nid,
        eracer.get_info(b) AS bi,
        eracer.get_info(m) AS mi,
        eracer.get_info(d) AS di
      FROM current
      LIMIT ALL -- prevent redundant get_info calls
    ) AS n ON o.ind_id = n.nid;

COMMENT ON VIEW output
IS 'compares predictions with original values';

--------------------------------------------------------------------------------

CREATE OR REPLACE VIEW summary AS
  SELECT
    sum(((bi).round>0)::int)::int AS b_upd,
    sum(((mi).round>0)::int)::int AS m_upd,
    sum(((di).round>0)::int)::int AS d_upd,
    sum(((bi).suspect)::int)::int AS b_sus,
    sum(((mi).suspect)::int)::int AS m_sus,
    sum(((di).suspect)::int)::int AS d_sus
  FROM (
    SELECT nid,
      eracer.get_info(b) AS bi,
      eracer.get_info(m) AS mi,
      eracer.get_info(d) AS di
    FROM current
    LIMIT ALL -- prevent redundant get_info calls
  ) AS q1;

COMMENT ON VIEW summary
IS 'breakdown of update and suspect counts';

--------------------------------------------------------------------------------

CREATE OR REPLACE VIEW status AS
  SELECT
    sum((NOT b_has AND b_rnd = 0)::int)::int AS b_mis,
    sum((NOT m_has AND m_rnd = 0)::int)::int AS m_mis,
    sum((NOT d_has AND d_rnd = 0)::int)::int AS d_mis,
    -- still corrupt == false negatives
    sum((b_has AND b_init <> b_orig AND NOT b_susp)::int)::int AS b_cor,
    sum((m_has AND m_init <> m_orig AND NOT m_susp)::int)::int AS m_cor,
    sum((d_has AND d_init <> d_orig AND NOT d_susp)::int)::int AS d_cor
  FROM (
    SELECT * FROM output
  ) AS q1;

COMMENT ON VIEW status
IS 'how many nodes are still missing or corrupt';

--------------------------------------------------------------------------------

CREATE OR REPLACE VIEW res_aq AS
  SELECT
    count(b_err)::integer AS b_cnt,
    avg(b_err)::real AS b_mae,
    avg(b_bas)::real AS b_bas,
    count(d_err)::integer AS d_cnt,
    avg(d_err)::real AS d_mae,
    avg(d_bas)::real AS d_bas
  FROM (
    SELECT -- absolute error; use mean prediction by default
      abs(COALESCE(b_exp, 1774) - b_orig) AS b_err,
      abs(COALESCE(d_exp, 1807) - d_orig) AS d_err,
      -- baseline absolute error using the mean prediction
      abs(1774 - b_orig) AS b_bas,
      abs(1807 - d_orig) AS d_bas
    FROM output
    WHERE NOT (b_has OR d_has) -- NOT present
  ) AS q1;

COMMENT ON VIEW res_aq
IS 'aggregate results: accuracy/quality (error)';

--------------------------------------------------------------------------------

CREATE OR REPLACE VIEW res_dc AS
  SELECT
    sum((b_has AND     b_init <> b_orig AND     b_susp)::int)::int AS b_tp,
    sum((b_has AND NOT b_init <> b_orig AND     b_susp)::int)::int AS b_fp,
    sum((b_has AND     b_init <> b_orig AND NOT b_susp)::int)::int AS b_fn,
    sum((b_has AND NOT b_init <> b_orig AND NOT b_susp)::int)::int AS b_tn,
    -- cannot tell which of the "init = orig" values are actually corrupt
    sum((d_has AND     d_init <> d_orig AND     d_susp)::int)::int AS d_tp,
    sum((d_has AND NOT d_init <> d_orig AND     d_susp)::int)::int AS d_fp,
    sum((d_has AND     d_init <> d_orig AND NOT d_susp)::int)::int AS d_fn,
    sum((d_has AND NOT d_init <> d_orig AND NOT d_susp)::int)::int AS d_tn
  FROM (
    SELECT * FROM output
    WHERE (b_has OR d_has) -- present
  ) AS q1;

COMMENT ON VIEW res_dc
IS 'aggregate results: data cleaning (tp & fn)';
