--
-- views for generating experiments and evaluating results
--

CREATE SCHEMA v1;
SET search_path TO v1;

CREATE OR REPLACE VIEW params AS
  SELECT 0.00 AS missing, 0.00 AS corrupt;

    COMMENT ON VIEW params
    IS 'missing/corrupt percentages for input view';

CREATE OR REPLACE VIEW current AS
  SELECT * FROM eracer.node;

    COMMENT ON VIEW current
    IS 'alias for node table after current round';

--------------------------------------------------------------------------------

CREATE OR REPLACE VIEW input AS
  SELECT nid,
    -- missing values almost always coincided in original data
    CASE WHEN missing THEN eracer.new_basis(NULL)
         WHEN corrupt THEN eracer.new_basis((15 + random() * 20)::real)
         ELSE t END AS t,
    CASE WHEN missing THEN eracer.new_basis(NULL)
         WHEN corrupt THEN eracer.new_basis((15 + random() * 45)::real)
         ELSE h END AS h,
    l, -- light was unpredictable in original data
    v -- voltage was always present in original data
  FROM (
    SELECT node2.*,
      random <= missing AS missing,
      random > missing AND random <= missing + corrupt AS corrupt
    FROM params,
      (SELECT *, random() FROM eracer.node) AS node2
  ) AS assign;

COMMENT ON VIEW input
IS 'generates random missing/corrupt data';

--------------------------------------------------------------------------------

CREATE OR REPLACE VIEW output AS
  SELECT n.nid,
    o.temperature AS t_orig,
      (ti).initval AS t_init,
      (ti).hasinit AS t_has,
      (ti).suspect AS t_susp,
      (ti).round AS t_rnd,
      (ti).expect AS t_exp,
      (ti).stddev AS t_std,
    o.humidity AS h_orig,
      (hi).initval AS h_init,
      (hi).hasinit AS h_has,
      (hi).suspect AS h_susp,
      (hi).round AS h_rnd,
      (hi).expect AS h_exp,
      (hi).stddev AS h_std,
    o.light AS l_orig,
      (li).initval AS l_init,
      (li).hasinit AS l_has,
      (li).suspect AS l_susp,
      (li).round AS l_rnd,
      (li).expect AS l_exp,
      (li).stddev AS l_std,
    o.voltage AS v_orig,
      (vi).initval AS v_init,
      (vi).hasinit AS v_has,
      (vi).suspect AS v_susp,
      (vi).round AS v_rnd,
      (vi).expect AS v_exp,
      (vi).stddev AS v_std
  FROM public.clean AS o
    INNER JOIN (
      SELECT nid,
        eracer.get_info(t) AS ti,
        eracer.get_info(h) AS hi,
        eracer.get_info(l) AS li,
        eracer.get_info(v) AS vi
      FROM current
      LIMIT ALL -- prevent redundant get_info calls
    ) AS n ON eracer.make_nid(o.second, o.moteid) = n.nid;

COMMENT ON VIEW output
IS 'compares predictions with original values';

--------------------------------------------------------------------------------

CREATE OR REPLACE VIEW summary AS
  SELECT
    sum(((ti).round>0)::int)::int AS t_upd,
    sum(((hi).round>0)::int)::int AS h_upd,
    sum(((li).round>0)::int)::int AS l_upd,
    sum(((vi).round>0)::int)::int AS v_upd,
    sum(((ti).suspect)::int)::int AS t_sus,
    sum(((hi).suspect)::int)::int AS h_sus,
    sum(((li).suspect)::int)::int AS l_sus,
    sum(((vi).suspect)::int)::int AS v_sus
  FROM (
    SELECT nid,
      eracer.get_info(t) AS ti,
      eracer.get_info(h) AS hi,
      eracer.get_info(l) AS li,
      eracer.get_info(v) AS vi
    FROM current
    LIMIT ALL -- prevent redundant get_info calls
  ) AS q1;

COMMENT ON VIEW summary
IS 'breakdown of update and suspect counts';

--------------------------------------------------------------------------------

CREATE OR REPLACE VIEW status AS
  SELECT
    sum((NOT t_has AND t_rnd = 0)::int)::int AS t_mis,
    sum((NOT h_has AND h_rnd = 0)::int)::int AS h_mis,
    sum((NOT l_has AND l_rnd = 0)::int)::int AS l_mis,
    sum((NOT v_has AND v_rnd = 0)::int)::int AS v_mis,
    -- still corrupt == false negatives
    sum((t_has AND t_init <> t_orig AND NOT t_susp)::int)::int AS t_cor,
    sum((h_has AND h_init <> h_orig AND NOT h_susp)::int)::int AS h_cor,
    sum((l_has AND l_init <> l_orig AND NOT l_susp)::int)::int AS l_cor,
    sum((v_has AND v_init <> v_orig AND NOT v_susp)::int)::int AS v_cor
  FROM (
    SELECT * FROM output
  ) AS q1;

COMMENT ON VIEW status
IS 'how many nodes are still missing or corrupt';

--------------------------------------------------------------------------------

CREATE OR REPLACE VIEW res_aq AS
  SELECT
    count(t_err)::integer AS t_cnt,
    avg(t_err)::real AS t_mae,
    avg(t_bas)::real AS t_bas,
    count(h_err)::integer AS h_cnt,
    avg(h_err)::real AS h_mae,
    avg(h_bas)::real AS h_bas
  FROM (
    SELECT -- absolute error; use mean prediction by default
      abs(COALESCE(t_exp, 22.1161) - t_orig) AS t_err,
      abs(COALESCE(h_exp, 39.5494) - h_orig) AS h_err,
      -- baseline absolute error using the mean prediction
      abs(22.1161 - t_orig) AS t_bas,
      abs(39.5494 - h_orig) AS h_bas
    FROM output
    WHERE NOT (t_has OR h_has) -- NOT present
  ) AS q1;

COMMENT ON VIEW res_aq
IS 'aggregate results: accuracy/quality (error)';

--------------------------------------------------------------------------------

CREATE OR REPLACE VIEW res_dc AS
  SELECT
    sum((t_has AND     t_init <> t_orig AND     t_susp)::int)::int AS t_tp,
    sum((t_has AND NOT t_init <> t_orig AND     t_susp)::int)::int AS t_fp,
    sum((t_has AND     t_init <> t_orig AND NOT t_susp)::int)::int AS t_fn,
    sum((t_has AND NOT t_init <> t_orig AND NOT t_susp)::int)::int AS t_tn,
    -- cannot tell which of the "init = orig" values are actually corrupt
    sum((h_has AND     h_init <> h_orig AND     h_susp)::int)::int AS h_tp,
    sum((h_has AND NOT h_init <> h_orig AND     h_susp)::int)::int AS h_fp,
    sum((h_has AND     h_init <> h_orig AND NOT h_susp)::int)::int AS h_fn,
    sum((h_has AND NOT h_init <> h_orig AND NOT h_susp)::int)::int AS h_tn
  FROM (
    SELECT * FROM output
    WHERE (t_has OR h_has) -- present
  ) AS q1;

COMMENT ON VIEW res_dc
IS 'aggregate results: data cleaning (tp & fn)';
