#
# visualize correlations between all pairs of attributes
# sample.csv + plots.R -> *.png + plots.tex -> plots.pdf
#

require(quantreg)  # optional

d <- read.csv("models/sample.csv")
vars = names(d)
nvar = length(vars)

# optional argument: number of attributes to infer
atts <- as.integer(commandArgs(TRUE)[1])
if (is.na(atts)) {
  atts <- nvar/2  # assuming one type of neighbors
}

# model visualization
lm.viz <- function(xlab, ylab, outdir=NULL) {
  # isn't there a better way to pass names and values?
  X <- eval(as.name(xlab), envir=d)
  Y <- eval(as.name(ylab), envir=d)
  print(paste(ylab, xlab, sep=" ~ "))

  # simple linear regression
  fm <- lm(Y ~ X)
  a <- coef(fm)[2]
  b <- coef(fm)[1]
  s <- sd(fm$residuals)
  r <- cor(Y, X, use="c")

  # quantile regression (optional)
  if (exists("rq")) {
    tau <- c(0.05, 0.25, 0.50, 0.75, 0.95)
    fm <- rq(Y ~ X, tau=tau, method="fn")
    qa <- coef(fm)[2,]
    qb <- coef(fm)[1,]
  }

  # generate the plot (optional: save to png)
  if (!is.null(outdir)) {
    png(filename=paste(outdir, "/", xlab, "_", ylab, ".png", sep=""),
        width=480, height=480, pointsize=12)
  }
  main <- paste(
      "a =", round(a, digits=2),
      ", b =", round(b, digits=2),
      ", sd =", round(s, digits=2),
      ", cc =", round(r, digits=2))
  plot(X, Y, xlab=xlab, ylab=ylab, main=main, col="gray")
  # regression model in green; +/- 2*sd in red
  abline(b, a, col="green", lwd=2)
  abline(b+2*s, a, col="red")
  abline(b-2*s, a, col="red")
  # quantile regression models in blue
  if (exists("rq")) {
    for (i in 1:length(tau)) {
      abline(qb[i], qa[i], col="blue")
    }
  }
  if (!is.null(outdir)) {
    dev.off()
  }
}

# intra-tuple correlations
for (i in seq(1, atts-1)) {
  for (j in seq(i+1, atts)) {
    lm.viz(vars[i], vars[j], "plots/intra")
  }
}

# inter-tuple correlations
for (i in seq(1, atts)) {
  for (j in seq(atts+1, nvar)) {
    lm.viz(vars[i], vars[j], "plots/inter")
  }
}
