"""
Removes models that will never be selected (e.g., hidden by more general ones).
"""

class Model:
    """One row of the models table."""

    def __init__(self, line):
        self.line = line
        self.toks = line.split(',')
        self.var = int(self.toks[0])
        self.rs = float(self.toks[2])
        self.se = float(self.toks[3])
        # models to prune by default
        if self.rs < 0.8 or line.find("NA") > -1:
            self.keep = False
            return
        for c in ('l', 'v', 'm'):  # ignored attributes
            if self.toks[1].find(c) > -1:
                self.keep = False
                return
        # parse and count the nonzero coefficients
        self.args = [float(arg) for arg in self.toks[6:]]
        self.narg = 0
        for a in self.args:
            if a != 0:
                # prune models with unessential parameters
                if abs(a) < 0.001:
                    self.keep = False
                    return
                self.narg += 1
        self.keep = True

    def __str__(self):
        return self.toks[1]  # model formula

    def __cmp__(self, next):
        """Comparison function (ORDER BY var, rs DESC, se)."""
        if self.var < next.var: return -1
        if self.var > next.var: return 1
        if self.rs < next.rs: return 1
        if self.rs > next.rs: return -1
        if self.se < next.se: return -1
        if self.se > next.se: return 1
        return 0

    def subset(self, next):
        """Is this model a subset of a subsequent one?"""
        # assuming: self.var == next.var
        for i in range(len(self.args)):
            if self.args[i] != 0 and next.args[i] == 0:
                return False
        return True


def fewer(mods):
    """Favor models with similar stats but fewer variables."""
    for i in reversed(range(len(mods))):
        if mods[i].keep:
            for j in reversed(range(i)):
                if mods[i].var != mods[j].var:
                    break # sorted by var
                # discard better models if not worth the cost
                if mods[j].keep and mods[i].subset(mods[j]):
                    rs2 = mods[j].rs - mods[i].rs
                    se2 = mods[i].se - mods[j].se
                    nv2 = mods[j].narg - mods[i].narg
                    # additional variables should improve the rs or se
                    if rs2 < nv2 * 0.02 and se2 < nv2 * 2.00:
                        mods[j].keep = False


def hidden(mods):
    """Remove models that will never be selected."""
    for i in range(len(mods)):
        if mods[i].keep:
            for j in range(i + 1, len(mods)):
                if mods[i].var != mods[j].var:
                    break # sorted by var
                # determine which models to delete
                if mods[j].keep and mods[i].subset(mods[j]):
                    mods[j].keep = False
                    #print "{0}\t{1}".format(mods[i], mods[j])


def main():
    # read the models into memory
    fin = open("models/models.csv", 'r')
    data = fin.readlines()
    fin.close()
    mods = [Model(line) for line in data[1:]]
    print "%d\t%s" % (len(mods), "from models.csv")
    # apply filters to models
    mods = filter(lambda m: m.keep, mods)
    print "%d\t%s" % (len(mods), "good rs & args")
    mods.sort()
    fewer(mods)
    mods = filter(lambda m: m.keep, mods)
    print "%d\t%s" % (len(mods), "worth the cost")
    hidden(mods)
    mods = filter(lambda m: m.keep, mods)
    print "%d\t%s" % (len(mods), "removed hidden")
    # export the resulting models
    fout = open("models/pruned.csv", 'w')
    fout.write(data[0])
    for mod in mods:
        fout.write(mod.line)
    fout.close()


if __name__ == "__main__":
    main()
