/*
 * model.c - initialization and model table sync
 */

#include "postgres.h"
#include "fmgr.h"
#include "executor/spi.h"
#include "model.h"

int32 g_nvar;
int32 g_nmod;
Model g_mods[MAX_MODS];

int16 g_round;
int32 g_updcnt;
int32 g_version;

/*----------------------------------------------------------------------------*\
  Initialization
\*----------------------------------------------------------------------------*/

/*
 * Ensures that the library is dynamically loaded into a compatible server.
 */
PG_MODULE_MAGIC;

/*
 * Called by the backend immediately after loading the library.
 */
void _PG_init(void) {
	c_model_sync(0);
	g_round = 0;
	g_updcnt = 0;
	g_version = 7;
}

/*----------------------------------------------------------------------------*\
  C Implementation
\*----------------------------------------------------------------------------*/

/*
 * Loads the model table into local memory. The optional parameter 'ext' allows
 * for multiple model tables, but this feature is deprecated.
 */
int32 c_model_sync(char ext) {
	int32 i, j;
	bool isnull;

	/* construct the applicable queries */
	const char *chk = "SELECT 1 FROM pg_tables "
			"WHERE schemaname = 'eracer' AND tablename = 'model'";
	static char sql[] = "SELECT * FROM eracer.model_? ORDER BY rs DESC, se";
	if (ext == 0) {
		sql[26] = ' ';
		sql[27] = ' ';
	} else {
		sql[26] = '_';
		sql[27] = ext;
	}

	/* clear previous version of models */
	g_nvar = 0;
	g_nmod = 0;
	MemSet(g_mods, 0, sizeof(g_mods));

	/* abort if the default model table doesn't exist */
	SPI_connect();
	if (ext == 0) {
		i = SPI_execute(chk, true, 0);
		if (i == SPI_OK_SELECT && SPI_processed == 0) {
			SPI_finish();
			return g_nmod;
		}
	}

	/* use SPI to retrieve the model table */
	i = SPI_execute(sql, true, 0);
	if (i == SPI_OK_SELECT && SPI_processed > 0) {
		const int32 B = 6; /* number of original columns; see tables.sql */
		TupleDesc tupdesc;

		/* validate the number of cols/rows */
		tupdesc = SPI_tuptable->tupdesc;
		if (tupdesc->natts - B > 2 * MAX_VARS) {
			ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
							errmsg("need to increase MAX_VARS in model.h")));
		}
		if (SPI_processed > MAX_MODS) {
			ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
							errmsg("need to increase MAX_MODS in model.h")));
		}
		g_nvar = (tupdesc->natts - B) / 2;
		g_nmod = SPI_processed;

		/* for each model row */
		for (i = 0; i < g_nmod; i++) {
			HeapTuple tuple = SPI_tuptable->vals[i];
			Model *mod = &g_mods[i];

			/* extract the header columns */
			mod->var = DatumGetInt32(SPI_getbinval(tuple, tupdesc, 1, &isnull));
			mod->var--; /* convert to zero-based index */
			strcpy(mod->form, SPI_getvalue(tuple, tupdesc, 2));
			mod->rs = DatumGetFloat4(SPI_getbinval(tuple, tupdesc, 3, &isnull));
			mod->se = DatumGetFloat4(SPI_getbinval(tuple, tupdesc, 4, &isnull));
			mod->df = DatumGetInt32(SPI_getbinval(tuple, tupdesc, 5, &isnull));
			mod->b0 = DatumGetFloat4(SPI_getbinval(tuple, tupdesc, 6, &isnull));

			/* extract the coefficients */
			for (j = 0; j < g_nvar; j++) {
				mod->ci[j] = DatumGetFloat4(SPI_getbinval(tuple, tupdesc,
						j + 7, &isnull));
				mod->cn[j] = DatumGetFloat4(SPI_getbinval(tuple, tupdesc,
						j + 7 + g_nvar, &isnull));
			}
		}
	}
	SPI_finish();
	return g_nmod;
}

/*----------------------------------------------------------------------------*\
  Postgres V1 Function Prototypes
\*----------------------------------------------------------------------------*/

Datum model_sync(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(model_sync);

Datum set_round(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(set_round);

Datum set_version(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(set_version);

/*----------------------------------------------------------------------------*\
  SQL Interface (Wrappers)
\*----------------------------------------------------------------------------*/

Datum model_sync(PG_FUNCTION_ARGS) {
	int32 ret;
	BpChar *bpc;
	char ext = 0;
	if (fcinfo->nargs > 0 && !PG_ARGISNULL(0)) {
		bpc = PG_GETARG_BPCHAR_P(0);
		ext = bpc->vl_dat[0];
	}
	ret = c_model_sync(ext);
	PG_RETURN_INT32(ret);
}

/*
 * Updates the current round number and returns/clears the update count.
 */
Datum set_round(PG_FUNCTION_ARGS) {
	int32 ret = g_updcnt;
	g_round = PG_GETARG_INT16(0);
	g_updcnt = 0;
	PG_RETURN_INT32(ret);
}

/*
 * Selects which variant of the ERACER algorithm to use (for the experiments).
 * Note that each feature is a power of two; they can be combined by addition.
 *
 * 1 : enable cleaning (replace)
 * 2 : don't propagate outliers
 * 4 : feedback control (distill)
 */
Datum set_version(PG_FUNCTION_ARGS) {
	int32 v = PG_GETARG_INT32(0);
	if (v < 0 || v > 7) {
		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
						errmsg("invalid version number (see model.c)")));
	}
	g_version = v;
	PG_RETURN_INT32(g_version);
}
