CREATE FUNCTION make_nid(integer, integer)
RETURNS integer AS $$
  -- use 22 bits for i and 6 bits for j; leave 4 bits for an optional k
  SELECT (($1 & 4194303) << 10) | (($2 & 63) << 4);
$$ LANGUAGE SQL IMMUTABLE STRICT;

CREATE FUNCTION nid_i(nid integer)
RETURNS integer AS $$
  SELECT ($1 >> 10) & 4194303;
$$ LANGUAGE SQL IMMUTABLE STRICT;

CREATE FUNCTION nid_j(nid integer)
RETURNS integer AS $$
  SELECT ($1 >> 4) & 63;
$$ LANGUAGE SQL IMMUTABLE STRICT;

CREATE FUNCTION nid_k(nid integer)
RETURNS integer AS $$
  SELECT $1 & 15;
$$ LANGUAGE SQL IMMUTABLE STRICT;

COMMENT ON FUNCTION make_nid(integer, integer) IS 'nid bitwise constructor';
COMMENT ON FUNCTION nid_i(nid integer) IS 'gets 1st component of nid';
COMMENT ON FUNCTION nid_j(nid integer) IS 'gets 2nd component of nid';
COMMENT ON FUNCTION nid_k(nid integer) IS 'gets 3rd component of nid';

--------------------------------------------------------------------------------

CREATE FUNCTION nodelt(node, node)
RETURNS boolean AS $$
  SELECT $1.nid < $2.nid;
$$ LANGUAGE SQL IMMUTABLE STRICT;

CREATE FUNCTION nodele(node, node)
RETURNS boolean AS $$
  SELECT $1.nid <= $2.nid;
$$ LANGUAGE SQL IMMUTABLE STRICT;

CREATE FUNCTION nodeeq(node, node)
RETURNS boolean AS $$
  SELECT $1.nid = $2.nid;
$$ LANGUAGE SQL IMMUTABLE STRICT;

CREATE FUNCTION nodege(node, node)
RETURNS boolean AS $$
  SELECT $1.nid >= $2.nid;
$$ LANGUAGE SQL IMMUTABLE STRICT;

CREATE FUNCTION nodegt(node, node)
RETURNS boolean AS $$
  SELECT $1.nid > $2.nid;
$$ LANGUAGE SQL IMMUTABLE STRICT;

CREATE FUNCTION nodecmp(node, node)
RETURNS integer AS $$
  -- ($1.nid - $2.nid) may overflow
  SELECT CASE
    WHEN $1.nid > $2.nid THEN 1
    WHEN $1.nid = $2.nid THEN 0
    ELSE -1 END;
$$ LANGUAGE SQL IMMUTABLE STRICT;

CREATE FUNCTION nodehash(node)
RETURNS integer AS $$
  SELECT hashint4($1.nid);
$$ LANGUAGE SQL IMMUTABLE STRICT;

COMMENT ON FUNCTION nodelt(node, node) IS '< operator function';
COMMENT ON FUNCTION nodele(node, node) IS '<= operator function';
COMMENT ON FUNCTION nodeeq(node, node) IS '= operator function';
COMMENT ON FUNCTION nodege(node, node) IS '>= operator function';
COMMENT ON FUNCTION nodegt(node, node) IS '> operator function';

COMMENT ON FUNCTION nodecmp(node, node) IS 'node comparison function';
COMMENT ON FUNCTION nodehash(node) IS 'node hash function';

--------------------------------------------------------------------------------

CREATE OPERATOR < (
  PROCEDURE = nodelt,
  LEFTARG = node,
  RIGHTARG = node,
  COMMUTATOR = >,
  RESTRICT = scalarltsel,
  JOIN = scalarltjoinsel
);

CREATE OPERATOR <= (
  PROCEDURE = nodele,
  LEFTARG = node,
  RIGHTARG = node,
  COMMUTATOR = >=,
  RESTRICT = scalarltsel,
  JOIN = scalarltjoinsel
);

CREATE OPERATOR = (
  PROCEDURE = nodeeq,
  LEFTARG = node,
  RIGHTARG = node,
  COMMUTATOR = =,
  RESTRICT = eqsel,
  JOIN = eqjoinsel,
  HASHES,
  MERGES
);

CREATE OPERATOR >= (
  PROCEDURE = nodege,
  LEFTARG = node,
  RIGHTARG = node,
  COMMUTATOR = <=,
  RESTRICT = scalargtsel,
  JOIN = scalargtjoinsel
);

CREATE OPERATOR > (
  PROCEDURE = nodegt,
  LEFTARG = node,
  RIGHTARG = node,
  COMMUTATOR = <,
  RESTRICT = scalargtsel,
  JOIN = scalargtjoinsel
);

--------------------------------------------------------------------------------

CREATE OPERATOR CLASS node_ops
DEFAULT FOR TYPE node USING btree AS
  OPERATOR 1 <  ,
  OPERATOR 2 <= ,
  OPERATOR 3 =  ,
  OPERATOR 4 >= ,
  OPERATOR 5 >  ,
  FUNCTION 1 nodecmp(node, node) ;

CREATE OPERATOR CLASS node_ops
DEFAULT FOR TYPE node USING hash AS
   OPERATOR 1 = ,
   FUNCTION 1 nodehash(node) ;
