/*
 * stats.c - statistical functions for approximate inference
 */

#include "postgres.h"
#include "funcapi.h"
#include "access/heapam.h"
#include "basis.h"
#include "model.h"

/* set to 1 to send debug output to the client */
#define DEBUG_PRINT 0
#if DEBUG_PRINT == 1
#define DPRINTF(format, args...) elog(NOTICE, format, args)
#else
#define DPRINTF(format, args...) ((void) 0)
#endif

/*
 * Internal state: first we save the nid, previous values, and basis vars from
 * the self tuple, and then store the running totals for the others. (Postgres
 * requires the data to be serialized in case aggregation spills to disk.)
 */
typedef struct {

	/* varlena header (do not touch directly!) */
	int32 vl_len_;

	/* nid of the self tuple */
	int32 nid;

	/* previous values of self (for distill) */
	float4 pvals[MAX_VARS];
	bool pnull[MAX_VARS];

	/* sums of expected values (rows are dst vars, cols are src vars) */
	float4 vi2[MAX_VARS][MAX_VARS];
	float4 vn2[MAX_VARS][MAX_VARS];

	/* counts of contributing values (for averaging and model selection) */
	int16 ni2[MAX_VARS][MAX_VARS];
	int16 nn2[MAX_VARS][MAX_VARS];

	/* Basis[] for self follows this struct */

} State;

static float4 prev_value(int32 dst, Basis *basis, bool *isnull);
static Basis *distill(State *s, int32 src, Basis *basis, bool itoi, int32 dst);

/*----------------------------------------------------------------------------*\
  Aggregate Transition
\*----------------------------------------------------------------------------*/

/*
 * Adds the expected value of the given basis to the state (via vm & cm).
 */
static void add_value(State *state, int32 src, Basis *basis, bool itoi) {
	int32 dst;
	Basis *upd;
	float4 value;
	bool isnull;

	/* get the target matrices, for simplicity */
	float4 (*vm)[MAX_VARS];
	int16 (*cm)[MAX_VARS];
	if (itoi) {
		vm = state->vi2;
		cm = state->ni2;
	} else {
		vm = state->vn2;
		cm = state->nn2;
	}

	/* feedback control: "filter out" influence from each variable */
	if (g_version & 4 && basis->pdf.model > -1) {
		DPRINTF("curr(%c[%d]): %g", itoi?'i':'n', src + 1,
				c_expect(&basis->pdf, &isnull));
		for (dst = 0; dst < g_nvar; dst++) {
			upd = distill(state, src, basis, itoi, dst);
			if (g_version & 2 && upd->suspect) {
				isnull = true; /* suppress outliers */
			} else {
				value = c_expect(&upd->pdf, &isnull);
			}
			if (!isnull) {
				vm[dst][src] += value;
				cm[dst][src] += 1;
			}
			if (upd != basis) {
				pfree(upd); /* allocated by distill */
				DPRINTF("distill(%d): %g", dst + 1, value);
			}
		}
	} else {
		/* propagate the same message to all variables */
		if (g_version & 2 && basis->suspect) {
			isnull = true; /* suppress outliers */
		} else {
			value = c_expect(&basis->pdf, &isnull);
		}
		if (!isnull) {
			for (dst = 0; dst < g_nvar; dst++) {
				vm[dst][src] += value;
				cm[dst][src] += 1;
			}
		}
	}
}

/*
 * Transition function; adds nodes to the internal state.
 */
static State *c_erace_trans(State *state, HeapTupleHeader self,
		HeapTupleHeader next) {
	int32 size, i;
	bool isnull;
	Basis *vars[MAX_VARS];
	char *ptr;

	/* if starting a new group */
	if (state == NULL) {
		if (HeapTupleHeaderGetNatts(self) != 1 + g_nvar) {
			ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
							errmsg("erace_trans() - node and model tables must "
									"define the same number of variables")));
		}

		/* extract attributes and calculate the state size */
		size = sizeof(State);
		for (i = 0; i < g_nvar; i++) {
			vars[i] = DatumGetBasisP(GetAttributeByNum(self, 2 + i, &isnull));
			size += VARSIZE(vars[i]);
		}

		/* allocate and initialize the state (once per group) */
		state = palloc(size);
		MemSet(state, 0, sizeof(State));
		SET_VARSIZE(state, size);
		state->nid = DatumGetInt32(GetAttributeByNum(self, 1, &isnull));
		ptr = ((char *) state) + sizeof(State);

		/* extract the previous values (if applicable) */
		if (g_version & 4) {
			for (i = 0; i < g_nvar; i++) {
				state->pvals[i] = prev_value(i, vars[i], &state->pnull[i]);
			}
		}

		/* compute expected values; append original vars */
		for (i = 0; i < g_nvar; i++) {
			add_value(state, i, vars[i], true);
			size = VARSIZE(vars[i]);
			memcpy(ptr, vars[i], size);
			ptr += size;
		}
	}

	/* extract and add the next set of expected values */
	if (next != NULL) {
		for (i = 0; i < g_nvar; i++) {
			vars[i] = DatumGetBasisP(GetAttributeByNum(next, 2 + i, &isnull));
			add_value(state, i, vars[i], false);
		}
	}
	return state;
}

/*----------------------------------------------------------------------------*\
  Aggregate Finalization
\*----------------------------------------------------------------------------*/

/*
 * Returns the first model (and its index) for which there is enough data.
 */
static Model *select_model(int32 dst, int16 ni[], int16 nn[], int16 *out) {
	int m, i;
	Model *model;
	for (m = 0; m < g_nmod; m++) {
		model = &g_mods[m];
		if (model->var == dst) {
			/* is there a value for every nonzero coefficient? */
			for (i = 0; i < g_nvar; i++) {
				if (model->ci[i] != 0 && ni[i] == 0) {
					break;
				}
				if (model->cn[i] != 0 && nn[i] == 0) {
					break;
				}
			}
			if (i == g_nvar) {
				DPRINTF("model = %s", model->call);
				*out = m;
				return model;
			}
		}
	}
	DPRINTF("%s", "no model found");
	return NULL;
}

/*
 * Is the original data outside the inferred pdf?
 */
static bool outlier(PDF *known, PDF *infer) {
	float4 a, b, c, d;
	c_confint(known, ERROR_THOLD, &a, &b);
	c_confint(infer, ERROR_THOLD, &c, &d);

	/* inferred range doesn't contain known */
	return (a < c || b > d);
}

/*
 * Compares the distributions before/after inference to identify errors in base
 * data and detect convergence. Returns the former basis with no correction, or
 * constructs/copies a modified version and increments the global update count.
 */
static Basis *evaluate(Basis *basis, PDF *pdf) {
	Basis *ret = basis; /* default: no change */
	bool missing = (basis->hasinit == false);

	/* validate existing data (identify errors) */
	if (g_version & 1 && !missing) {
		char buf[NORM_BUF_SIZE];
		PDF *init = get_init_pdf(basis, buf);
		if (!basis->suspect) {
			/* error correction take 1: check for suspicious base data */
			if (outlier(init, pdf) == true) {
				DPRINTF("%s", "follow suspect");
				ret = update_basis(basis, pdf);
				ret->suspect = true;
			}
		} else {
			/* error correction take 2: check for premature suspicion */
			if (outlier(init, pdf) == false) {
				DPRINTF("%s", "cancel suspect");
				ret = revert_basis(basis);
				ret->suspect = false;
			} else {
				/* treat suspicious base data as if it were missing */
				missing = true;
			}
		}
	}

	/* test for convergence (missing or suspicious) */
	if (missing) {
		if (basis->pdf.model != pdf->model) {
			DPRINTF("%s", "model changed");
			ret = update_basis(basis, pdf);
		} else if (c_diverge(&basis->pdf, pdf) > DELTA_THOLD) {
			DPRINTF("%s", "pdf diverges");
			ret = update_basis(basis, pdf);
		}
	}
	return ret;
}

/*
 * The mother of all erace functions.
 * Note that var is 0-indexed in C, but is 1-indexed in SQL.
 */
static Basis *erace_var(int32 dst, Basis *basis, float4 vi[], float4 vn[],
		int16 ni[], int16 nn[]) {
	Model *model;
	int32 i;
	char *ptr;
	char buf[NORM_BUF_SIZE + MAX_VARS * 12] = {0};
	PDF *pdf = (PDF *) buf;
	DPRINTF("erace_var(var=%d)", dst + 1);

	/* is there a model for the observed values? */
	model = select_model(dst, ni, nn, &pdf->model);
	if (model == NULL) {
		return NULL;
	}

	/* apply the model; construct the pdf */
	pdf->type = PDF_NORM;
	pdf->length = 2;
	pdf->data[0] = model->b0;
	for (i = 0; i < g_nvar; i++) {
		pdf->data[0] += model->ci[i] * vi[i];
		pdf->data[0] += model->cn[i] * vn[i];
	}
	pdf->data[1] = model->se;
	DPRINTF("predict = %g, stddev = %g", pdf->data[0], pdf->data[1]);

	/* append predictors; evaluate changes */
	if (g_version & 4) {
		ptr = (char *) &pdf->data[pdf->length];
		memcpy(ptr, vi, g_nvar * 4);
		ptr += g_nvar * 4;
		memcpy(ptr, vn, g_nvar * 4);
		ptr += g_nvar * 4;
		memcpy(ptr, ni, g_nvar * 2);
		ptr += g_nvar * 2;
		memcpy(ptr, nn, g_nvar * 2);
	}
	return evaluate(basis, pdf);
}

/*
 * Finalize function; performs inference for each variable.
 */
static HeapTupleHeader c_erace_final(State *state, TupleDesc tupdesc) {
	int32 i, j;
	Basis *basis;
	HeapTuple tuple;
	Datum values[1 + MAX_VARS];
	bool isnull[1 + MAX_VARS] = {0};
	char *ptr = ((char *) state) + sizeof(State);
	DPRINTF("erace_final(nid=%d)", state->nid);

	/* compute averages of the neighbor values */
	for (i = 0; i < g_nvar; i++) {
		DPRINTF("averaging var %d", i + 1);
		for (j = 0; j < g_nvar; j++) {
			DPRINTF("i[%d] = %g x %d", j+1, state->vi2[i][j], state->ni2[i][j]);
			if (state->nn2[i][j] > 0) {
				state->vn2[i][j] /= state->nn2[i][j];
			}
			DPRINTF("n[%d] = %g x %d", j+1, state->vn2[i][j], state->nn2[i][j]);
		}
	}

	/* run inference on each var and construct the result */
	values[0] = Int32GetDatum(state->nid);
	for (i = 0; i < g_nvar; i++) {
		basis = erace_var(i, (Basis *) ptr, state->vi2[i], state->vn2[i],
				state->ni2[i], state->nn2[i]);
		if (basis == NULL) {
			/* no model found; no change */
			basis = (Basis *) ptr;
		}
		values[1 + i] = PointerGetDatum(basis);
		ptr += VARSIZE(ptr);
	}
	tuple = heap_form_tuple(tupdesc, values, isnull);
	return tuple->t_data;
}

/*----------------------------------------------------------------------------*\
  Feedback Control
\*----------------------------------------------------------------------------*/

/*
 * Gets the previous expected value (if any) for the given basis, i.e., the
 * "non-distilled" message sent during the last round.
 */
static float4 prev_value(int32 dst, Basis *basis, bool *isnull) {
	float4 *vi;
	int16 *ni;
	float4 value;
	PDF *pdf = &basis->pdf;
	if (pdf->model > -1) {
		/* get the (unused) message sent to itself */
		vi = &pdf->data[pdf->length];
		ni = (int16 *) (vi + g_nvar * 2);
		value = vi[dst];
		*isnull = (ni[dst] == 0);
	} else {
		/* no previous message; value was the same */
		value = c_expect(pdf, isnull);
	}
	return value;
}

/*
 * Filters out the contribution (if any) of the given dst variable.
 */
static Basis *distill(State *s, int32 src, Basis *basis, bool itoi, int32 dst) {
	float4 vz[MAX_VARS]; /* 'z' means one of the values has been zeroed out */
	int16 nz[MAX_VARS];
	Basis *upd;
	PDF *pdf = &basis->pdf;
	Model *model = &g_mods[pdf->model];

	/* get the predictor values (basis is read-only) */
	float4 *vi = &pdf->data[pdf->length];
	float4 *vn = vi + g_nvar;
	int16 *ni = (int16 *) (vn + g_nvar);
	int16 *nn = ni + g_nvar;

	/* for correctness, check if dst wasn't used to infer src */
	if (s->pnull[dst]) {
		return basis;
	}
	if (itoi) {
		if (model->ci[dst] == 0) {
			return basis;
		}
		memcpy(vz, vi, g_nvar * 4);
		memcpy(nz, ni, g_nvar * 2);
	} else {
		if (model->cn[dst] == 0) {
			return basis;
		}
		memcpy(vz, vn, g_nvar * 4);
		memcpy(nz, nn, g_nvar * 2);
	}

	/* re-compute the predictor as if dst's value had been missing */
	if (nz[dst] > 1) {
		vz[dst] = (vz[dst] * nz[dst] - s->pvals[dst]) / (nz[dst] - 1);
		nz[dst] = (nz[dst] - 1);
	} else {
		vz[dst] = 0;
		nz[dst] = 0;
	}

	/* redo erace of src with the updated observed values */
	if (itoi) {
		upd = erace_var(src, basis, vz, vn, nz, nn);
	} else {
		upd = erace_var(src, basis, vi, vz, ni, nz);
	}
	if (upd == NULL) {
		/* no model found; use initial pdf */
		upd = revert_basis(basis);
		upd->suspect = false;
	}
	return upd;
}

/*----------------------------------------------------------------------------*\
  Postgres V1 Function Prototypes
\*----------------------------------------------------------------------------*/

Datum erace_trans(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(erace_trans);

Datum erace_final(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(erace_final);

/*----------------------------------------------------------------------------*\
  SQL Interface (Wrappers)
\*----------------------------------------------------------------------------*/

Datum erace_trans(PG_FUNCTION_ARGS) {
	State *state;
	HeapTupleHeader self;
	HeapTupleHeader next;
	if (!fcinfo->context || !IsA(fcinfo->context, AggState)) {
		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
						errmsg("erace_trans() - must call from aggregate")));
	}
	if (PG_ARGISNULL(0)) {
		state = NULL; /* new group */
	} else {
		state = (State *) PG_GETARG_BYTEA_P(0);
	}
	if (PG_ARGISNULL(1)) {
		ereport(ERROR, (errcode(ERRCODE_NULL_VALUE_NOT_ALLOWED),
						errmsg("erace_trans() - self node cannot be NULL")));
	}
	self = PG_GETARG_HEAPTUPLEHEADER(1);
	if (PG_ARGISNULL(2)) {
		next = NULL; /* no neighbors */
	} else {
		next = PG_GETARG_HEAPTUPLEHEADER(2);
	}
	state = c_erace_trans(state, self, next);
	PG_RETURN_BYTEA_P(state);
}

Datum erace_final(PG_FUNCTION_ARGS) {
	TupleDesc tupdesc;
	State *state;
	HeapTupleHeader tuple;
	if (!fcinfo->context || !IsA(fcinfo->context, AggState)) {
		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
						errmsg("erace_final() - must call from aggregate")));
	}
	if (get_call_result_type(fcinfo, NULL, &tupdesc) != TYPEFUNC_COMPOSITE) {
		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
						errmsg("function returning record called in context "
								"that cannot accept type record")));
	}
	tupdesc = BlessTupleDesc(tupdesc);
	state = (State *) PG_GETARG_BYTEA_P(0);
	tuple = c_erace_final(state, tupdesc);
	PG_RETURN_HEAPTUPLEHEADER(tuple);
}
